/*
 * Copyright © 2020 Fachhochschule für die Wirtschaft (FHDW) Hannover
 *
 * This file is part of gaming-core.
 *
 * Gaming-core is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Gaming-core is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * gaming-core. If not, see <http://www.gnu.org/licenses/>.
 */
package de.fhdw.gaming.core.domain;

import java.util.Map;
import java.util.Optional;

/**
 * Represents a game.
 *
 * @param <P>  The type of game players.
 * @param <S>  The type of game states.
 * @param <M>  The type of game moves.
 * @param <ST> The type of game strategies.
 */
public interface Game<P extends Player<P>, S extends State<P, S>, M extends Move<P, S>, ST extends Strategy<P, S, M>>
        extends AutoCloseable {

    /**
     * Returns an integer value that uniquely identifies this game.
     */
    int getId();

    /**
     * Returns all the players of the game.
     */
    Map<String, P> getPlayers();

    /**
     * Returns all the players' strategies.
     */
    Map<String, ST> getStrategies();

    /**
     * Returns a copy of the game state.
     */
    S getState();

    /**
     * Adds an observer.
     *
     * @param observer The observer to add.
     */
    void addObserver(Observer observer);

    /**
     * Removes an observer.
     *
     * @param observer The observer to remove.
     */
    void removeObserver(Observer observer);

    /**
     * Starts the game.
     *
     * @throws InterruptedException if starting the game has been interrupted.
     */
    void start() throws InterruptedException;

    /**
     * Makes a single move. This means:
     * <ul>
     * <li>The set of all players being able to make a move is determined. If this set is empty, the game is over.</li>
     * <li>The strategies of all these players are requested to choose a move.</li>
     * <li>The first such move is applied to the game state (if possible, as the move may be illegal).</li>
     * <li>The game state is checked whether the game is over. This is typically the case if:
     * <ul>
     * <li>One or more players have won the game; or</li>
     * <li>All players have lost the game; or</li>
     * <li>All but one player have lost the game. In that case the remaining player has won the game.</li>
     * </ul>
     * </li>
     * </ul>
     * <p>
     * If some player's strategy returns {@code null} or an illegal move, this player loses the game immediately.
     * <p>
     * The following events are possibly emitted:
     * <ul>
     * <li>{@link Observer#nextPlayersComputed(Game, State, java.util.Set)}</li>
     * <li>{@link Observer#illegalPlayerRejected(Game, State, Player)}</li>
     * <li>{@link Observer#legalMoveApplied(Game, State, Player, Move)}</li>
     * <li>{@link Observer#illegalMoveRejected(Game, State, Player, Optional, String)}</li>
     * <li>{@link Observer#playerOvertaken(Game, State, Player, Player)}</li>
     * <li>{@link Observer#finished(Game, State)}</li>
     * </ul>
     *
     * @throws IllegalStateException if the game is already over.
     * @throws InterruptedException  if choosing and applying a move has been interrupted.
     */
    void makeMove() throws InterruptedException;

    /**
     * Tries to abort the game by invoking {@link Strategy#abortRequested(int)} on all players' strategies.
     */
    void abortRequested();

    /**
     * Chooses a random but valid move. Used when a strategy does not return a move in the given time window.
     *
     * @param player The player to choose the move for.
     * @param state  The game state. This is a copy of the game state.
     * @return A valid move. Can be empty if no valid move is available (e.g. because some player has already won the
     *         game).
     */
    Optional<M> chooseRandomMove(P player, S state);

    /**
     * Determines whether the game has been started.
     *
     * @return {@code true} if the game has been started, {@code false} otherwise.
     */
    boolean isStarted();

    /**
     * Determines whether the game is over.
     *
     * @return {@code true} if the game is over, {@code false} otherwise.
     */
    boolean isFinished();

    @Override
    void close();

}
