package de.fhdw.gaming.core.domain.util;

import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import de.fhdw.gaming.core.domain.Game;
import de.fhdw.gaming.core.domain.GameBuilder;
import de.fhdw.gaming.core.domain.GameBuilderFactory;
import de.fhdw.gaming.core.domain.GameException;
import de.fhdw.gaming.core.domain.Move;
import de.fhdw.gaming.core.domain.Player;
import de.fhdw.gaming.core.domain.State;
import de.fhdw.gaming.core.domain.Strategy;
import de.fhdw.gaming.core.ui.InputProvider;
import de.fhdw.gaming.core.ui.InputProviderException;
import de.fhdw.gaming.core.ui.util.ChainedInputProvider;
import de.fhdw.gaming.core.ui.util.NonInteractiveInputProvider;

/**
 * Helper facade for creating a default {@link Game} given a {@link GameBuilderFactory}.
 */
public final class GameBuilderFacade {

    /**
     * Private constructor.
     */
    private GameBuilderFacade() {
        // nothing to do
    }

    /**
     * Creates a {@link GameBuilder} given the desired type of a {@link GameBuilderFactory}.
     *
     * @param <GB>                      The desired type of a {@link GameBuilder}.
     * @param <GBF>                     The desired type of a {@link GameBuilderFactory}.
     * @param <P>                       The expected type of a {@link Player}.
     * @param <S>                       The expected type of a {@link State}.
     * @param <M>                       The expected type of a {@link Move}.
     * @param <ST>                      The expected type of a {@link Strategy}.
     * @param type                      The desired {@link Class} of the {@link GameBuilder} to return.
     * @param gameBuilderFactories      An iterator over all available {@link GameBuilderFactory} objects.
     * @param strategy                  A {@link Strategy} to use for all players. If empty, the
     *                                  {@link GameBuilderFactory} is asked for all available strategies, and the first
     *                                  available one is used for the players.
     * @param maxComputationTimePerMove The maximum number of seconds a strategy can use for choosing a move.
     * @return The {@link GameBuilder} or {@link Optional#empty()} if either {@link GameBuilderFactory game builder
     *         factories} or {@link Strategy strategies} are unavailable.
     */
    public static <GB extends GameBuilder, GBF extends GameBuilderFactory, P extends Player<P>, S extends State<P, S>,
            M extends Move<P, S>, ST extends Strategy<P, S, M>> Optional<GB> createGameBuilder(final Class<GB> type,
                    final Iterable<GBF> gameBuilderFactories, final Optional<ST> strategy,
                    final int maxComputationTimePerMove)
                    throws InputProviderException, GameException {

        final Iterator<GBF> gbfIter = gameBuilderFactories.iterator();
        if (!gbfIter.hasNext()) {
            return Optional.empty();
        }

        final GBF gameBuilderFactory = gbfIter.next();

        final ST gameStrategy;
        if (strategy.isPresent()) {
            gameStrategy = strategy.orElseThrow();
        } else {
            @SuppressWarnings("unchecked")
            final List<ST> strategies = (List<ST>) gameBuilderFactory.getStrategies();
            if (strategies.isEmpty()) {
                return Optional.empty();
            }
            gameStrategy = strategies.get(0);
        }

        return Optional.of(createGame(type, gameBuilderFactory, gameStrategy, maxComputationTimePerMove));
    }

    /**
     * Creates a {@link GameBuilder} from a {@link GameBuilderFactory} and an optional {@link Strategy}.
     *
     * @param <GB>                      The desired type of a {@link GameBuilder}.
     * @param <GBF>                     The type of a {@link GameBuilderFactory}.
     * @param <P>                       The expected type of a {@link Player}.
     * @param <S>                       The expected type of a {@link State}.
     * @param <M>                       The expected type of a {@link Move}.
     * @param <ST>                      The expected type of a {@link Strategy}.
     * @param type                      The desired {@link Class} of the {@link GameBuilder} to return.
     * @param gameBuilderFactory        The {@link GameBuilderFactory} to use.
     * @param strategy                  The {@link Strategy} to use. If empty, the first available one is used.
     * @param maxComputationTimePerMove The maximum number of seconds a strategy can use for choosing a move.
     * @return The {@link GameBuilder} or {@link Optional#empty()} if either {@link GameBuilderFactory game builder
     *         factories} or {@link Strategy strategies} are unavailable.
     */
    private static <GB extends GameBuilder, GBF extends GameBuilderFactory, P extends Player<P>, S extends State<P, S>,
            M extends Move<P, S>, ST extends Strategy<P, S, M>> GB createGame(final Class<GB> type,
                    final GBF gameBuilderFactory, final ST strategy, final int maxComputationTimePerMove)
                    throws InputProviderException, GameException {

        InputProvider mostRecentProvider = null;
        int nextPlayerId = gameBuilderFactory.getMinimumNumberOfPlayers();
        while (nextPlayerId > 0) {
            final InputProvider lastPlayerProvider = mostRecentProvider;
            final InputProvider playerInputProvider = lastPlayerProvider == null ? new NonInteractiveInputProvider()
                    : new ChainedInputProvider(
                            new NonInteractiveInputProvider(),
                            (final Map<String, Object> lastDataSet) -> lastPlayerProvider);

            playerInputProvider.fixedString(GameBuilderFactory.PARAM_PLAYER_NAME, Integer.toString(nextPlayerId--));
            playerInputProvider.fixedObject(GameBuilderFactory.PARAM_PLAYER_STRATEGY, strategy);
            mostRecentProvider = playerInputProvider;
        }

        final InputProvider firstPlayerProvider = mostRecentProvider;
        final InputProvider gameInputProvider = firstPlayerProvider == null ? new NonInteractiveInputProvider()
                : new ChainedInputProvider(
                        new NonInteractiveInputProvider(),
                        (final Map<String, Object> lastDataSet) -> firstPlayerProvider);
        gameInputProvider
                .fixedInteger(GameBuilderFactory.PARAM_MAX_COMPUTATION_TIME_PER_MOVE, maxComputationTimePerMove);

        return type.cast(gameBuilderFactory.createGameBuilder(gameInputProvider));
    }
}
