/*
 * Copyright © 2020 Fachhochschule für die Wirtschaft (FHDW) Hannover
 *
 * This file is part of gaming-core.
 *
 * Gaming-core is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Gaming-core is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * gaming-core. If not, see <http://www.gnu.org/licenses/>.
 */
package de.fhdw.gaming.core.ui;

import java.util.Map;
import java.util.Optional;
import java.util.Set;

import de.fhdw.gaming.core.ui.type.validator.Validator;

/**
 * Provides user input by creating a dialogue with the user. First, a sequence of fields is described, then the contents
 * of these fields are requested from the user. The form of presentation is decided by the concrete implementation. For
 * example, this may be done by prompting the fields one after another on the text console, or by showing a GUI dialog
 * where the user may type the data in arbitrary order and submit all data at once.
 */
public interface InputProvider {

    /**
     * Indicates the need for a string value from the user.
     *
     * @param id           The ID of the field.
     * @param prompt       The prompt.
     * @param defaultValue An optional default value.
     * @param validators   Additional validators to use.
     * @return {@code this}
     */
    @SuppressWarnings("unchecked")
    InputProvider needString(String id, String prompt, Optional<String> defaultValue,
            Validator<String>... validators) throws InputProviderException;

    /**
     * Adds a predefined string value. Requires that {@link #needString(String, String, Optional, Validator...)} is
     * called for the same field before requesting data from the user, otherwise calling this operation has no effect.
     *
     * @param id         The ID of the field.
     * @param fixedValue The corresponding value.
     * @return {@code this}
     */
    InputProvider fixedString(String id, String fixedValue) throws InputProviderException;

    /**
     * Indicates the need for an integer value from the user.
     *
     * @param id           The ID of the field.
     * @param prompt       The prompt.
     * @param defaultValue An optional default value.
     * @param validators   Additional validators to use.
     * @return {@code this}
     */
    @SuppressWarnings("unchecked")
    InputProvider needInteger(String id, String prompt, Optional<Integer> defaultValue,
            Validator<Integer>... validators) throws InputProviderException;

    /**
     * Adds a predefined integer value. Requires that {@link #needInteger(String, String, Optional, Validator...)}
     * is called for the same field before requesting data from the user, otherwise calling this operation has no
     * effect.
     *
     * @param id         The ID of the field.
     * @param fixedValue The corresponding value.
     * @return {@code this}
     */
    InputProvider fixedInteger(String id, Integer fixedValue) throws InputProviderException;

    /**
     * Indicates the need for a boolean value from the user.
     *
     * @param id           The ID of the field.
     * @param prompt       The prompt.
     * @param defaultValue An optional default value.
     * @param validators   Additional validators to use.
     * @return {@code this}
     */
    @SuppressWarnings("unchecked")
    InputProvider needBoolean(String id, String prompt, Optional<Boolean> defaultValue,
            Validator<Boolean>... validators) throws InputProviderException;

    /**
     * Adds a predefined boolean value. Requires that {@link #needBoolean(String, String, Optional, Validator...)}
     * is called for the same field before requesting data from the user, otherwise calling this operation has no
     * effect.
     *
     * @param id         The ID of the field.
     * @param fixedValue The corresponding value.
     * @return {@code this}
     */
    InputProvider fixedBoolean(String id, Boolean fixedValue) throws InputProviderException;

    /**
     * Indicates the need for one of multiple objects from the user.
     *
     * @param id           The ID of the field.
     * @param prompt       The prompt.
     * @param defaultValue An optional default value.
     * @param objectSet    The values to choose from. May be arbitrary objects; the mapping is established through their
     *                     string values retrieved by calling {@link Object#toString()}
     * @return {@code this}
     */
    InputProvider needObject(String id, String prompt, Optional<Object> defaultValue, Set<?> objectSet)
            throws InputProviderException;

    /**
     * Adds a predefined object reference. Requires that {@link #needObject(String, String, Optional, Set)} is called
     * for the same field before requesting data from the user, otherwise calling this operation has no effect.
     *
     * @param id         The ID of the field.
     * @param fixedValue The corresponding value.
     * @return {@code this}
     */
    InputProvider fixedObject(String id, Object fixedValue) throws InputProviderException;

    /**
     * Requests the data from the user.
     *
     * @param title Some title for the user interface.
     * @return The data mapping field IDs to their contents.
     * @throws InputProviderException if the operation has been aborted prematurely (e.g. if the user cancelled a
     *                                dialog).
     */
    Map<String, Object> requestData(String title) throws InputProviderException;

    /**
     * Creates a new instance of this input provider for requesting the next set of data from the user.
     *
     * @param lastDataSet The data requested the last time. May be empty.
     * @throws InputProviderException if some input is not compatible.
     */
    InputProvider getNext(Map<String, Object> lastDataSet) throws InputProviderException;
}
